package Beispiele::iX::POP3Server;

use constant CRLF => "\x0d\x0a";
use Apache2::Connection;
use Apache2::Const -compile => qw(OK DECLINED);
use Apache2::Filter;
use Apache2::HookRun;
use Apache2::RequestRec;
use Apache2::RequestUtil;
use APR::Bucket;
use APR::Brigade;
use APR::Const -compile => qw(EOF SO_NONBLOCK SUCCESS);
use APR::Error;
use APR::Socket;

use Beispiele::iX::POP3;
use Beispiele::iX::Mbox;


sub handler
{   my $c = shift;

    # --- Requestrecord anlegen ---
    my $r = Apache2::RequestRec->new($c);
    $r->location_merge(__PACKAGE__);
    $r->filename(__PACKAGE__);
    $r->uri(__PACKAGE__);

    # --- Zugriffskontrolle ---
    my $acc = $r->run_access_checker();
    return $acc unless Apache2::Const::OK == $acc
        || Apache2::Const::DECLINED == $acc;

    # --- Socket Optionen setzen ---
    $c->client_socket()->opt_set(APR::Const::SO_NONBLOCK => 0);

    # --- Bucket Brigades anlegen ---
    my $bbIn  = APR::Brigade->new($c->pool(), $c->bucket_alloc());
    my $bbOut = APR::Brigade->new($c->pool(), $c->bucket_alloc());

    # --- zustzliche Konfig lesen ---
    my $mboxDir = $r->dir_config("mboxdir");

    # --- POP-Objekt initialisieren ---
    my $last = 0;
    my ($pop, $greeting) = new Beispiele::iX::POP3
    (   mbox  => sub
        {   my $user = shift;
            return new Beispiele::iX::Mbox("$mboxDir/$user");
        }
    ,   quit  => sub {$last = 1}
    ,   basic => sub
        {   my ($user, $pass) = @_;
            $r->set_basic_credentials($user, $pass);
            my $auth = $r->run_check_user_id();

            return Apache2::Const::OK == $auth
                || Apache2::Const::DECLINED == $auth;
        }
    );
    output($c, $bbOut, $greeting);

    # --- IO-Schleife ---
    while (defined(my $line = getline($c, $bbIn)))
    {   output($c, $bbOut, $pop->client($line));
        last if $last;
    }

    # --- Brigaden verwerfen ---
    $bbIn->destroy();
    $bbOut->destroy();

    # --- Zugriff protokollieren ---
    $r->run_log_transaction();

    # --- Handler beenden ---
    return Apache2::Const::OK;
}


# -----  getline()  -----
# * liest Zeilen von der Bucket Brigade; entfernt CRLF am Ende
#
sub getline
{   my ($c, $bb) = @_;

    my $buffer = "";
    while (1)
    {   # --- Bucket Brigade befllen ---
        my $rc = $c->input_filters()
            ->get_brigade($bb, Apache2::Const::MODE_GETLINE);

        return if $rc == APR::Const::EOF;
        die APR::Error::strerror($rc) unless $rc == APR::Const::SUCCESS;

        # --- Brigade auslesen, CRLF suchen und entfernen ---
        my $data = "";
        while (!$bb->is_empty())
        {   my $b = $bb->first();
            last if $b->is_eos();   # "end of stream"
            $b->remove();
            $b->read($data);
            $buffer .= $data;
            my $crlf = index $buffer, CRLF;
            return substr $buffer, 0, $crlf if $crlf > -1;
        }
    }
}


# -----  output()  -----
# * schreibt Daten auf Bucket Brigade
#
sub output
{   my ($c, $bb, $data) = @_;

    # --- Netz-Zeilenenden ---
    $data =~ s/\n/CRLF/eg;

    # --- Bucket anlegen, befllen und an Brigade anfgen ---
    my $bucket = APR::Bucket->new($bb->bucket_alloc(), $data);
    $bb->insert_tail($bucket);

    # --- Brigade auf Ausgabefilter schreiben ---
    my $rc = $c->output_filters->fflush($bb);
    die APR::Error::strerror($rc) unless $rc == APR::Const::SUCCESS;
}


1;
